import time
import os
import cohere

from llamp.llms.base_llm_system import BaseLLMSystem

# from tenacity import (
#     retry,
#     stop_after_attempt, # type: ignore
#     wait_random_exponential, # type: ignore
# )

class CohereText(BaseLLMSystem):
    def __init__(self, system_name="CohereText",save_path="game_logs", temperature = 0.0, model="command", stop_sequences=None):
        
        super().__init__(system_name, save_path, temperature=temperature)        
        self.base_prompt = [{
            "role" : "system",
            "content" : "You will interact with the environment to solve the given task. Think step by step "
        }]
        self.reset()
        API_KEY = os.environ['COHERE_API_KEY']
        self.co = cohere.Client(API_KEY)

        self.temperature = temperature
        self.model = model
        self.stop_sequences = stop_sequences

    # @retry(wait=wait_random_exponential(min=1, max=60), stop=stop_after_attempt(6),reraise=True)
    def call_model(self, temperature=None):
        """Call OpenAI API"""

        prompt = self.generate_text_prompt()

        response = self.co.generate(
            model=self.model,
            prompt = prompt,
            temperature=self.temperature,
            end_sequences=self.stop_sequences #to have same behaviour as in Openai
            # stop_sequences=["}\n"] #included in text end_sequences if excluded
        )



        # print("="*20)
        # print(response)
        # print(type(response))
        # input(">")
        # print(response.generations[0].text)
        # input(">>")

        answer = response.generations[0].text

        return answer


if __name__=="__main__":
    print("Nothing to run here.")



 


from llamp.llms.api.anthropic_chat import AnthropicChat
from llamp.llms.api.anthropic_text import AnthropicText

from llamp.llms.api.cohere_chat import CohereChat
from llamp.llms.api.cohere_text import CohereText
from llamp.llms.api.cohere_chat_text import CohereChatText

from llamp.llms.api.openai_chat import OpenAIChat
from llamp.llms.api.openai_text import OpenAIText
from llamp.llms.api.openai_chat_text import OpenAIChatText
from llamp.llms.api.openai_chat_text_sampling import OpenAIChatTextSampling

from llamp.llms.api.nvidia_chat_text import NvidiaChatText

from llamp.llms.api.cerebras_chat_text import CerebrasChatText



  
from llamp.llms import (
    Manual
)

if __name__=="__main__":
    agent = Manual()    
    
    print("\nTesting the act method:")   
    response, token_info = agent.act("Hi", return_token_count=True)
    print(f"\nResponse:\n{response}")
    print(f"\nResponse:\n{response}")
    print(f"\nToken Information:")
    print(f"Input Tokens: {token_info['in_token_all']}")
    print(f"Message Tokens: {token_info['in_token_message']}")
    print(f"Output Tokens: {token_info['out_token_action']}")



  
from llamp.llms.local import (
    VLLMChat
)

if __name__=="__main__":
    model = "Qwen/Qwen2.5-0.5B-Instruct"
    agent = VLLMChat(
        model=model,
        tensor_parallel_size=1,
        temperature=0.5
    )    
    
    print("\nTesting the act method:")   
    response, token_info = agent.act("Hi", return_token_count=True)
    print(f"\nResponse:\n{response}")
    print(f"\nToken Information:")
    print(f"Input Tokens: {token_info['in_token_all']}")
    print(f"Message Tokens: {token_info['in_token_message']}")
    print(f"Output Tokens: {token_info['out_token_action']}")
    # agent.save()



from datetime import datetime
import os
import json

import tiktoken

from llamp.llms.base_system import BaseSystem


class BaseLLMSystem(BaseSystem):
	def __init__(self, system_name="base_llm",save_path="game_logs", temperature=0):
		super().__init__(system_name, save_path)
		self.base_prompt=[]
		self.current_prompt = self.base_prompt
		self.system_name = system_name
		self.file_name = self.get_save_path(save_path)
		self.temperature=temperature
	
	def _extend_memory(self, memory):
		""" Helper function to extend memory"""
		memory.update({"temperature":self.temperature})
		return memory

	def _count_in_token(self, current_observation):
		""" counts in tokens based on current_observation. """
		tokens_until_now = self.count_tokens()
		current_message_tokens = self.count_tokens(current_observation)
		in_tokens = tokens_until_now+current_message_tokens
		
		additional_logging = {
			"in_token_all": in_tokens,
			"in_token_message" : current_message_tokens
		}

		return additional_logging

	def _count_out_token(self, action):
		""" Counts out token based on action. """
		out_token = self.count_tokens(action)
		new_additional_logging = {
			"out_token_action":out_token		
		}

		return new_additional_logging

	def act(self, current_observation, temperature=None, return_token_count=False):
		""" Acting"""
		additional_logging_in = self._count_in_token(current_observation)
		self.add_to_history(current_observation, "user", additional_logging_in)
		
		in_tokens_count = self.count_tokens()
		if not (additional_logging_in["in_token_all"]==in_tokens_count):
			print("WARNING: In Tokens need to be same length.")
			print(f"additional_logging_in={additional_logging_in['in_token_all']}")
			print(f"in_tokens_count={in_tokens_count}")

		action = self.call_model(temperature)
		additional_logging_out = self._count_out_token(action)

		action = self.post_process_model_output(action)
		self.add_to_history(action, "assistant", additional_logging_out)

		additional_logging = {}
		additional_logging.update(additional_logging_in)
		additional_logging.update(additional_logging_out)
		if return_token_count:
			return action, additional_logging
		else:
			return action

	def call_model(self, temperature=None, return_token_count=False):
		"""Not Implemented in base class."""
		raise NotImplementedError("call model needs to be implemented")

	def post_process_model_output(self,model_output):
		"""Some post processing"""
		return model_output
   	
	def generate_text_prompt(self):
		"""Generates a text prompt for the 'old school' LLMs."""
		prompt = ""
		for section in self.current_prompt:
			prompt +=  section["content"]
		# prompt+="\n"
		return prompt

	def count_tokens(self, optional_text=None, model_name="gpt-3.5"):
		"""
		By default we are returning the openai tokenizer count. 
		If a model wants a better tokenizer it needs to provide its own.
		"""
		model_name_clean = "-".join(model_name.split("-")[:2])
		encoding = tiktoken.encoding_for_model(model_name_clean)
		if optional_text:
			token_count = len(encoding.encode(optional_text))
		else:
			prompt = self.generate_text_prompt()
			token_count = len(encoding.encode(prompt))

		return token_count


if __name__=="__main__":
	agent = BaseLLMAgent()
	print("Nothing to run here.")
	agent.add_to_history("hi","user")
	agent.add_to_history("do this","assistant")
	print(agent.generate_text_prompt())




# 
# 	    	{
# 		    	"role": "system", 
# 		   		"content": 
# """
# You are a robot that is able to take decisions in an interactive environment. 
# The `user` will be the actual game environment telling you about what your observations from the environment.

# Your responses should follow the syntax of the game. 
# You should respond with one command at every interaction only. 
# Each command should be a single command from the list of valid commands. 
# This is the list of available and valid commands together with a short description of each command.
# <<<
# look:                describe the current room
# goal:                print the goal of this game
# inventory:           print player's inventory
# go <dir>:            move the player north, east, south or west
# examine ...:         examine something more closely
# eat ...:             eat edible food
# open ...:            open a door or a container
# close ...:           close a door or a container
# drop ...:            drop an object on the floor
# take ...:            take an object that is on the floor
# put ... in/on ...:      place an object on a supporter
# take ... from ...:   take an object from a container or a supporter
# insert ... into ...: place an object into a container
# lock ... with ...:   lock a door or a container with a key
# unlock ... with ...: unlock a door or a container with a key 
# >>>
# """
# 	    	}
	    



from datetime import datetime
import os
import json
from copy import deepcopy


class AnyDict:
    def __init__(self, value=-1):
        self.value = value

    def __getitem__(self,key):
        return self.value

class BaseSystem():
    LOG_FILE_ENDING = ".json"

    def __init__(self, system_name="base",save_path="game_logs", *args, **kwargs):
        self.base_prompt = []
        self.current_prompt = []
        self.full_history = []
        self.system_name = system_name
        self.save_path_base = save_path
        self.file_name = self.get_save_path()

    def call_model(self):
        """Not Implemented in base class."""
        raise NotImplementedError("call model needs to be implemented")


    def act(self, current_observation, return_token_count=False, *args, **kwargs):
        """ Acting"""
        self.add_to_history(current_observation, "user")

        action = self.call_model()

        self.add_to_history(action, "assistant")

        if return_token_count:
            return action, AnyDict()
        else:
            return action

    def reset(self, *args, **kwargs):
        """Resets the system (so far only the current_prompt)"""
        self.current_prompt = deepcopy(self.base_prompt)
        self.full_history = deepcopy(self.current_prompt)
        self.file_name = self.get_save_path()


    def _extend_memory(self, memory, *args, **kwargs):
        """Extends memory. Change for specific classes to record more data"""
        return memory

    def add_first_observation(self, first_obs, *args, **kwargs):
        """ First Observation """
        self.add_to_history(first_obs, "user")

    def add_to_history(self, content, role, additional_data={}, *args, **kwargs):
        """ 
        Adds to the prompt.
        IF ERROR: This function used to be called  `add_to_prompt`
        """
        tmp = {
            "role": role,
            "content": content
            }

        self.current_prompt.append(tmp)

        tmp.update(additional_data)
        
        # To log additional data by default
        tmp = self._extend_memory(tmp)
        self.full_history.append(tmp)


    def pop_from_history(self, return_full=True, *args, **kwargs):
        """Removes last element from history."""
        try:
            if return_full:
                self.current_prompt.pop()
                return self.full_history.pop()
            else:
                self.full_history.pop()
                return self.current_prompt.pop() 
        except IndexError as e:
            pass


    def update_latest_history(self, content, key="content", old_key="old_content", *args, **kwargs):
        """
        Updates the latest history with new content. By default uses the key 'content' 
        """
        previous_content = self.full_history[-1][key]
        self.full_history[-1][key] = content
        self.full_history[-1][old_key] = previous_content


    def _extract_prompt_from_history(self, history, keys_to_keep_in_prompt=["role","content"], *args, **kwargs):
        """Extracts prompt from history."""
        prompt = []
        for segment in history:
            tmp = {}
            for key in keys_to_keep_in_prompt:
                tmp[key] = segment.get(key)

            prompt.append(tmp)
        return prompt

    def load_from_saved_data(self, previous_history, keys_to_keep_in_prompt=["role","content"], *args, **kwargs):
        """
        Loads the saved prompt
        Input Options:
        - Previous History as list
        - Previous Histroy as FileName
        - Previous History as FilePointer
        """
        self.reset()
        
        import io
        if type(previous_history)==list:
            self.full_history = previous_history

        elif type(previous_history)==str:
            import json
            with open(previous_history) as file:
                previous_history = json.load(file)
            self.full_history = previous_history

        elif type(previous_history)==io.TextIOWrapper:
            previous_history = json.load(previous_history)
            self.full_history = previous_history

        self.current_prompt = self._extract_prompt_from_history(self.full_history, keys_to_keep_in_prompt)


    def set_base_prompt_and_reset(self,base_prompt, *args, **kwargs):
        """Sets a new base prompt and resets the agent."""
        self.base_prompt = base_prompt
        self.reset()

    
    def save(self, *args, **kwargs):
        """Saves game interaction to file"""
        with open(self.file_name,"w") as file:
            json.dump(self.full_history,file,indent=4)

    def get_file_name(self, base_name="logs", *args, **kwargs):
        """creates a file name based on datetime"""
        now = datetime.now()
        file_name = self.system_name+"_"+base_name+"_"+now.strftime("%d_%m_%Y_%H_%M_%S")+self.LOG_FILE_ENDING
        return file_name

    def get_save_path(self, save_path="", *args, **kwargs):
        """ Get Save path """
        if save_path:
            self.save_path_base = save_path
        self.create_save_path(self.save_path_base)
        file_name = os.path.join(self.save_path_base,self.get_file_name())
        return file_name

    def update_save_path(self,save_path, *args, **kwargs):
        """ Update Save Path """
        self.save_path_base = save_path
        self.file_name = self.get_save_path(self.save_path_base)

    @staticmethod
    def create_save_path(save_path, *args, **kwargs):
        """Creates the save path if it doesn't exist"""
        paths = os.path.split(save_path)
        current_path = ""
        for path in paths:
            if path: 
                current_path = os.path.join(current_path,path)
                if not os.path.exists(current_path):
                    os.mkdir(current_path)


    def count_tokens(self, *args, **kwargs):
        """ Creating this function for completeness """
        return 0

    def is_resample(self, *args, **kwargs):
        """Returns whether this system is designed for resampling."""
        return False
    




if __name__=="__main__":
    print("Nothing to run here.")



from llamp.llms.base_system import BaseSystem

class Manual(BaseSystem):
    def __init__(self, system_name="Manual",save_path="game_logs"):
        super().__init__(system_name, save_path)
   

    def call_model(self):
        """The action is just whatever the user inputs"""
        action = input("[[INPUT YOUR ACTION]] >>")
        return action     


if __name__=="__main__":
    agent = Manual()    
    
    print("\nTesting the act method:")   
    response, token_info = agent.act("Hi", return_token_count=True)
    print(f"\nResponse:\n{response}")
    print(f"\nToken Information:")
    print(f"Input Tokens: {token_info['in_token_all']}")
    print(f"Message Tokens: {token_info['in_token_message']}")
    print(f"Output Tokens: {token_info['out_token_action']}")


    response = agent.act("Hi", return_token_count=False)
    print(f"\nResponse:\n{response}")



import time
import os

import anthropic
from anthropic import HUMAN_PROMPT, AI_PROMPT

from llamp.llms.base_llm_system import BaseLLMSystem

# from tenacity import (
#     retry,
#     stop_after_attempt, # type: ignore
#     wait_random_exponential, # type: ignore
# )

class AnthropicChat(BaseLLMSystem):
	def __init__(self, system_name="AnthropicChat",save_path="game_logs", temperature=0.0, model="claude-2.1", stop_sequences=None):
		
		super().__init__(system_name, save_path, temperature=temperature)		
		self.client = anthropic.Anthropic(
		    # defaults to os.environ.get("ANTHROPIC_API_KEY")
		    # api_key="my_api_key",
		)
		# self.openai_attempts = 0

		self.temperature = temperature
		self.model = model
		self.stop_sequences = stop_sequences


	# @retry(wait=wait_random_exponential(min=1, max=60), stop=stop_after_attempt(6), reraise=True)
	def call_model(self):
		"""Call OpenAI API"""
		message = self.client.messages.create(
		    model=self.model,
		    max_tokens=1000,
		    temperature=self.temperature,
		    system=self.current_prompt[0]["content"],
		    messages=self.current_prompt[1:],
		    stop_sequences=self.stop_sequences
		)
		
		response = message.content[0].text

		return response




if __name__=="__main__":
	print("Nothing to run here.")
	# def call_model(self, attempt_limit = 100):
	# 	"""Call OpenAI API"""
	# 	try:
	# 		print("Trying to call GPT")
	# 		chat_completion = self.client.chat.completions.create(
	# 		    # model="gpt-3.5-turbo",
	# 		    model="gpt-3.5-turbo-0125",
	# 		    # model="gpt-4-turbo-preview",
	# 		    messages=self.current_prompt,
	# 		    temperature=0.8
	# 		)
	# 		chat_message = chat_completion.choices[0].message.content
	# 		self.openai_attempts = 0
	# 	except openai.RateLimitError:
	# 		print(f"Need to wait ... attempt:{self.openai_attempts}")
	# 		time.sleep(20)
	# 		self.openai_attempts+=1
	# 		if self.openai_attempts < attempt_limit:
	# 			return self.call_model()
	# 		else:
	# 			raise Exception("Exceeded Attempt Limit")

	# 	return chat_message


import torch
import transformers

from llamp.llms.base_llm_system import BaseLLMSystem

class Yi(BaseLLMSystem):
	def __init__(self, system_name="Yi",save_path="game_logs", test_mode=False, temperature=0.0):

		super().__init__(system_name, save_path, temperature=temperature)		
		if torch.cuda.is_available():
			torch.set_default_device("cuda")
		else:
			torch.set_default_device("cpu")

		if test_mode:
			self.base_prompt=[
				{
					"content": "You are an assistant that answers one word.",
					"role": "system"
				}
			]
			self.reset()

		self.model_name = "01-ai/Yi-6b-Chat-4bits"

		self.model = transformers.AutoModelForCausalLM.from_pretrained(self.model_name, device_map='auto', offload_folder="offload_yi")

		# https://github.com/huggingface/transformers/issues/27132
		# please use the slow tokenizer since fast and slow tokenizer produces different tokens
		self.tokenizer = transformers.AutoTokenizer.from_pretrained(
			self.model_name,
			use_fast=False,
		)		

	def construct_prompt_for_model(self):
	    """
	    Constructs a prompt for the ORCA 2 model using the OpenAI structure of `messages`
	    I.e. messages = [
	        {
	            "content" : "Bla bla bla",
	            "role": "assistant / user / system"
	        }
	    ]
	    """
	    prompt_decorators ={
	        "assistant" : {
	            "start" : "<|im_start|>assistant\n",
	            "end" : "<|im_end|>\n"
	        },
	        "user" : {
	            "start" : "<|im_start|>user\n",
	            "end" : "<|im_end|>\n"
	        },
	        "system" : {
	            "start" : "<|im_start|>system\n",
	            "end" : "<|im_end|>\n"
	        }        
	    }
	    prompt = ""
	    for message in self.current_prompt:
	        role = message["role"]
	        prompt += prompt_decorators[role]["start"]
	        prompt += message["content"]
	        prompt += prompt_decorators[role]["end"]

	    prompt += prompt_decorators["assistant"]["start"]

	    return prompt

	@staticmethod
	def extract_answer(output):
		"""Extracts the final answer from the Orca answer."""
		final_answer = output.split("assistant\n")[-1]
		final_answer = final_answer.replace("</s>","")
		return final_answer


	def call_model(self, attempt_limit = 100):
		"""Call OpenAI API"""
		
		prompt = self.construct_prompt_for_model()
		inputs = self.tokenizer(prompt, return_tensors='pt')
		output_ids = self.model.generate(inputs["input_ids"],)
		answer = self.tokenizer.batch_decode(output_ids)[0]

		final_answer = self.extract_answer(answer)

		return final_answer


if __name__=="__main__":
	print("Nothing to run here.")
